/*********************************************************************************
*                    "Die andere Uhr" mit LED-Matrix-8x8
*                    ===================================
*                             Uwe Berger; 2010
*
*
*
* ---------
* Have fun!
*
**********************************************************************************/
#ifndef F_CPU
	#define F_CPU 1000000
#endif

#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <util/delay.h>
#include <avr/pgmspace.h>
#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "font.h"

//*** Hardwaredefinitionen ***
// Matrix-Zeilen
#define ROW_0_DDR		DDRC
#define ROW_0_PORT		PORTC
#define ROW_0			PC0
#define ROW_1_DDR		DDRC
#define ROW_1_PORT		PORTC
#define ROW_1			PC1
#define ROW_2_DDR		DDRC
#define ROW_2_PORT		PORTC
#define ROW_2			PC2
#define ROW_3_DDR		DDRC
#define ROW_3_PORT		PORTC
#define ROW_3			PC3
#define ROW_4_DDR		DDRC
#define ROW_4_PORT		PORTC
#define ROW_4			PC4
#define ROW_5_DDR		DDRC
#define ROW_5_PORT		PORTC
#define ROW_5			PC5
#define ROW_6_DDR		DDRD
#define ROW_6_PORT		PORTD
#define ROW_6			PD0
#define ROW_7_DDR		DDRD
#define ROW_7_PORT		PORTD
#define ROW_7			PD1
// Matrix-Spalten
#define COL_0_DDR		DDRD
#define COL_0_PORT		PORTD
#define COL_0			PD3
#define COL_1_DDR		DDRD
#define COL_1_PORT		PORTD
#define COL_1			PD4
#define COL_2_DDR		DDRD
#define COL_2_PORT		PORTD
#define COL_2			PD5
#define COL_3_DDR		DDRD
#define COL_3_PORT		PORTD
#define COL_3			PD6
#define COL_4_DDR		DDRD
#define COL_4_PORT		PORTD
#define COL_4			PD7
#define COL_5_DDR		DDRB
#define COL_5_PORT		PORTB
#define COL_5			PB0
#define COL_6_DDR		DDRB
#define COL_6_PORT		PORTB
#define COL_6			PB1
#define COL_7_DDR		DDRB
#define COL_7_PORT		PORTB
#define COL_7			PB2

// Taster
#define TASTER_ON_OFF_DDR	DDRD
#define TASTER_ON_OFF_PORT	PORTD
#define TASTER_ON_OFF_PINX  PIND
#define TASTER_ON_OFF_PIN	PD2

#define TASTER_LEFT_DDR		DDRB
#define TASTER_LEFT_PORT	PORTB
#define TASTER_LEFT_PINX	PINB
#define TASTER_LEFT_PIN		PB7

#define TASTER_RIGHT_DDR	DDRB
#define TASTER_RIGHT_PORT	PORTB
#define TASTER_RIGHT_PINX	PINB
#define TASTER_RIGHT_PIN	PB6

// Tastencodes
#define ON_OFF	1		// Uhrzeit im Klartext
#define LEFT	2		// Stunden stellen
#define RIGHT	3		// Minuten stellen

#define ON		1
#define OFF		0 

#define DEBOUNCE_MS 100

#define PRELOAD_T0 	250
#define PRELOAD_T1 	49911

// ein paar Dimensionen
#define MATRIX_DX		8	// Breite LED-Matrix
#define MATRIX_DY		8	// Hoehe LED-Matrix

volatile uint8_t row = 0;
volatile uint8_t col = 0;

// Anzeigepuffer
volatile uint8_t led[8] = {
							0b00000000,
							0b00000000,
							0b00000000,
							0b00000000,
							0b00000000,
							0b00000000,
							0b00000000,
							0b00000000
							};
							
volatile uint8_t click;
uint8_t h, m, s, cs;
uint8_t old_s = 60;
uint8_t old_m = 60;
uint8_t old_h = 60;
							
							
//************************************************************************
void matrix_clear(void) {
	// alle Zeilen aus (Low)
	ROW_0_PORT &= ~(1<<ROW_0);
	ROW_1_PORT &= ~(1<<ROW_1);
	ROW_2_PORT &= ~(1<<ROW_2);
	ROW_3_PORT &= ~(1<<ROW_3);
	ROW_4_PORT &= ~(1<<ROW_4);
	ROW_5_PORT &= ~(1<<ROW_5);
	ROW_6_PORT &= ~(1<<ROW_6);
	ROW_7_PORT &= ~(1<<ROW_7);
	// alle Spalten aus (High)
	COL_0_PORT |= (1<<COL_0);
	COL_1_PORT |= (1<<COL_1);
	COL_2_PORT |= (1<<COL_2);
	COL_3_PORT |= (1<<COL_3);
	COL_4_PORT |= (1<<COL_4);
	COL_5_PORT |= (1<<COL_5);
	COL_6_PORT |= (1<<COL_6);
	COL_7_PORT |= (1<<COL_7);
}

//************************************************************************
uint8_t get_led(uint8_t x, uint8_t y)
{
	return (bit_is_set(led[y], 7-x));
}

//************************************************************************
ISR(TIMER1_OVF_vect)
{
	// Aufruf jede Sekunde --> Uhrentakt
	// ...
	TCNT1     = PRELOAD_T1;
	click = 1;
	// Uhrzeit weiterzaehlen
	s++;
	if (s > 59) {
		s = 0;
		m++;
		if (m > 59) {
			m = 0;
			h++;
			if (h > 23) {
				h = 0;
			}
		}
	}
}

//************************************************************************
//... jede LED einzeln angesteuert...
ISR(TIMER0_OVF_vect)
{
	TCNT0 = PRELOAD_T0;	
	// Matrix zuruecksetzen
	matrix_clear();
	// naechste Zeile/Spalte ermitteln
	row++;
	if (row > 7) {
		row = 0;
		col++;
		if (col > 7) col=0;
	}
	// naechste Spalte ein (Low)
	switch (col) {
		case 0: if (bit_is_set(led[row], MATRIX_DX-1-0)) COL_0_PORT &= ~(1<<COL_0); break;
		case 1: if (bit_is_set(led[row], MATRIX_DX-1-1)) COL_1_PORT &= ~(1<<COL_1); break;
		case 2: if (bit_is_set(led[row], MATRIX_DX-1-2)) COL_2_PORT &= ~(1<<COL_2); break;
		case 3: if (bit_is_set(led[row], MATRIX_DX-1-3)) COL_3_PORT &= ~(1<<COL_3); break;
		case 4: if (bit_is_set(led[row], MATRIX_DX-1-4)) COL_4_PORT &= ~(1<<COL_4); break;
		case 5: if (bit_is_set(led[row], MATRIX_DX-1-5)) COL_5_PORT &= ~(1<<COL_5); break;
		case 6: if (bit_is_set(led[row], MATRIX_DX-1-6)) COL_6_PORT &= ~(1<<COL_6); break;
		case 7: if (bit_is_set(led[row], MATRIX_DX-1-7)) COL_7_PORT &= ~(1<<COL_7); break;
	}
	// naechste Zeile ein (High)
	switch (row) {
		case 0: ROW_0_PORT |= (1 << ROW_0); break;
		case 1: ROW_1_PORT |= (1 << ROW_1); break;
		case 2: ROW_2_PORT |= (1 << ROW_2); break;
		case 3: ROW_3_PORT |= (1 << ROW_3); break;
		case 4: ROW_4_PORT |= (1 << ROW_4); break;
		case 5: ROW_5_PORT |= (1 << ROW_5); break;
		case 6: ROW_6_PORT |= (1 << ROW_6); break;
		case 7: ROW_7_PORT |= (1 << ROW_7); break;
	}
}

//************************************************************************
void my_delay_ms(uint16_t ms)
{
	uint16_t i;
	for (i=0; i<ms; i++) _delay_ms(1);
}

//************************************************************************
uint8_t scankey(volatile uint8_t *pinx, uint8_t pin) 
{
	if (bit_is_clear(*pinx, pin)) {
		my_delay_ms(DEBOUNCE_MS);
		if (bit_is_clear(*pinx, pin)) return 1;
	}
	return 0;
}

//************************************************************************
uint8_t getkey(void) 
{
	if (scankey(&TASTER_ON_OFF_PINX, TASTER_ON_OFF_PIN))	return ON_OFF;
	if (scankey(&TASTER_LEFT_PINX, TASTER_LEFT_PIN))		return LEFT;
	if (scankey(&TASTER_RIGHT_PINX, TASTER_RIGHT_PIN))		return RIGHT;
	return 0;
}

//************************************************************************
void clear_all(void)
{
	uint8_t i;
	for (i=0; i<8; i++) led[i] = 0;
}

//*********************************************************************
void set_led(uint8_t x, uint8_t y, uint8_t val)
{
	x = 7-x;
	if (val) {led[y] |= (1 << x);} else {led[y] &= ~(1 << x);}
}

//*********************************************************************
uint8_t my_rand(uint8_t basis)
{
	uint16_t x;
	while((x = rand()) >= RAND_MAX - (RAND_MAX % basis)); 
	return x % basis;
}

//*********************************************************************
// seed-Wert fuer srand() generieren; Quelle: 
// http://www.roboternetz.de/wissen/index.php/Zufallszahlen_mit_avr-gcc
unsigned int get_seed(void)
{
	uint16_t seed = 0;
	uint16_t *p = (uint16_t*) (RAMEND+1);
	extern uint16_t __heap_start;
	while (p >= &__heap_start + 1)
		seed ^= * (--p);
	return seed;
}

//************************************************************************
void set_sec (uint8_t s, uint8_t val) 
{
	switch (s) {
		case 0:  {set_led(0, 0, val); break;}
		case 1:  {set_led(1, 0, val); break;}
		case 2:  {set_led(2, 0, val); break;}
		case 3:  {set_led(3, 0, val); break;}
		case 4:  {set_led(4, 0, val); break;}
		case 5:  {set_led(5, 0, val); break;}
		case 6:  {set_led(6, 0, val); break;}
		case 7:  {set_led(7, 0, val); break;}
		case 8:  {set_led(7, 1, val); break;}
		case 9:  {set_led(7, 2, val); break;}
		case 10: {set_led(7, 3, val); break;}
		case 11: {set_led(7, 4, val); break;}
		case 12: {set_led(7, 5, val); break;}
		case 13: {set_led(7, 6, val); break;}
		case 14: {set_led(7, 7, val); break;}
		case 15: {set_led(6, 7, val); break;}
		case 16: {set_led(5, 7, val); break;}
		case 17: {set_led(4, 7, val); break;}
		case 18: {set_led(3, 7, val); break;}
		case 19: {set_led(2, 7, val); break;}
		case 20: {set_led(1, 7, val); break;}
		case 21: {set_led(0, 7, val); break;}
		case 22: {set_led(0, 6, val); break;}
		case 23: {set_led(0, 5, val); break;}
		case 24: {set_led(0, 4, val); break;}
		case 25: {set_led(0, 3, val); break;}
		case 26: {set_led(0, 2, val); break;}
		case 27: {set_led(0, 1, val); break;}
	}
}

//************************************************************************
void set_area(uint8_t x0, uint8_t y0, uint8_t x1, uint8_t y1, uint8_t count)
{
	uint8_t i, j;	
	uint8_t dx=x1-x0+1;
	uint8_t dy=y1-y0+1;
	uint8_t r=my_rand(9);
	for (j=y0; j<(y1+1); j++) {
		for (i=x0; i<(x1+1); i++) {
			set_led(i, j, OFF);
		}
	}
	// rein theoretisch koennte es hier zu einer Endlosschleife kommen,
	// die Wahrscheinlichkeit wird immer hoeher, je groesser count
	for (i=0; i<count; i++) {
		while (get_led(x0+(r%dx), y0+(r/dy))) r=my_rand(9);
		set_led(x0+(r%dx), y0+(r/dy), ON);
	}
}

//************************************************************************
void print_sign(uint8_t sign, int8_t x)
{
	uint8_t i, row;
	int8_t j;
	// anzeigbares Zeichen? Wenn nein, dann Leerzeichen!
	if ((sign < 0x20) || (sign > 0x7F)) sign = 0x20;
	// jede Zeile des Fonts adressieren, lesen, ausgeben
	for(i=0; i<FONT_DY; i++) {
		row = (pgm_read_byte(&font[(sign-0x20)*FONT_DY+i]));
		// erstmal soweit nach rechts (pos. x) oder links (neg. x) schieben
		if (x<0) {
			// links schieben
			for (j=0; j>x; j--) row *= 2;
		} else {
			// rechts schieben
			for (j=0; j<x; j++) row /= 2;
		}
		// und mit led[] verodern (damit schon dastehendes erhalten bleibt)
		led[i] |= row;
	}
}

//************************************************************************
void print_text(const char * text)
{
	uint8_t text_x_len = strlen(text)*(FONT_DX+1);
	uint8_t cursor = 0;
	uint8_t idx_sign; 
	int8_t mx;
	
	while (cursor<text_x_len) {
		clear_all();
		// ...erstes sichtbares Zeichen des Textes ermitteln
		idx_sign=cursor/(FONT_DX+1);
		// ...x-Anfangsposition des ersten noch sichtbaren Zeichen in der Matrix
		mx=0-(cursor%(FONT_DX+1));
		// ...und alles sichbare ausgeben
		while (mx<MATRIX_DX) {
			// ... irgendwie muss hier noch eine Pruefung rein ...!
			print_sign(text[idx_sign], mx);
			mx += FONT_DX+1;
			idx_sign++;
		}
		// ...Spaltencursor im Gesamttext erhoehen
		cursor++;
		// ...und etwas warten
		my_delay_ms(150);
	}
	clear_all();
}

//************************************************************************
//************************************************************************
//************************************************************************
int main(void)
{
	char txt[20];
	// Hardware initialisieren...
	// ... Analogcomperator ausschalten (Strom sparen!)
	ACSR = 0x80;		
	//... Ausgaenge
	ROW_0_DDR |= (1<<ROW_0);
	ROW_1_DDR |= (1<<ROW_1);
	ROW_2_DDR |= (1<<ROW_2);
	ROW_3_DDR |= (1<<ROW_3);
	ROW_4_DDR |= (1<<ROW_4);
	ROW_5_DDR |= (1<<ROW_5);
	ROW_6_DDR |= (1<<ROW_6);
	ROW_7_DDR |= (1<<ROW_7);
	COL_0_DDR |= (1<<COL_0);
	COL_1_DDR |= (1<<COL_1);
	COL_2_DDR |= (1<<COL_2);
	COL_3_DDR |= (1<<COL_3);
	COL_4_DDR |= (1<<COL_4);
	COL_5_DDR |= (1<<COL_5);
	COL_6_DDR |= (1<<COL_6);
	COL_7_DDR |= (1<<COL_7);
	// ... Eingaenge
	TASTER_ON_OFF_PORT &= ~(1 << TASTER_ON_OFF_PIN);
	TASTER_LEFT_PORT   &= ~(1 << TASTER_LEFT_PIN);
	TASTER_RIGHT_PORT  &= ~(1 << TASTER_RIGHT_PIN);
	// ... Timer0 initialisieren
    TCNT0   = PRELOAD_T0;
	TCCR0 |= (1 << CS01)|(1 << CS00);				// 64
	// ... Timer1 initialisieren
	TCNT1   = PRELOAD_T1;
	TCCR1B |= (1 << CS11) | (1 << CS10);			// 1024
	// ... Timermaske
	TIMSK  |= (1 << TOIE0) | (1 << TOIE1);			
	// ... Interrupts einschalten
	sei();
	// Zufallsgenerator initialisieren
	srand(get_seed());					
	while (1){
		// Tasten abfragen/reagieren
		switch (getkey()) {
			case ON_OFF:	{	// Uhrzeit als Lauftext ausgeben
								sprintf(txt, "%02d:%02d", h, m);
								print_text(txt);
								set_area(1, 1, 3, 3, h/10);
								set_area(4, 1, 6, 3, h%10);				
								set_area(1, 4, 3, 6, m/10);
								set_area(4, 4, 6, 6, m%10);
								break;
							}
			case LEFT:		{
								// Stunden stellen
								h++;
								if (h>23) h=0;	
								old_h=h;	
								// ... und anzeigen
								set_area(1, 1, 3, 3, h/10);
								set_area(4, 1, 6, 3, h%10);				
								break;
							}
			case RIGHT:		{
								// Minuten stellen
								m++;
								if (m>59) m=0;
								old_m=m;
								// ... und anzeigen
								set_area(1, 4, 3, 6, m/10);
								set_area(4, 4, 6, 6, m%10);
								break;
							}
		}
		// Uhr eine Sekunde weiter
		if (click) {
			click = 0;
			// Uhrzeit-Ausgabe
			// ... Sekunden
			// alte Sekunde loeschen
			set_sec(cs, OFF);
			cs++;
			// neue Sekunde setzen
			if (cs > 27) {cs = 0;}
			set_sec(cs, ON);
			// ... Minuten
			if (old_m != m) {
				old_m = m;
				// Minuten 10er
				set_area(1, 4, 3, 6, m/10);
				// Minuten 1er
				set_area(4, 4, 6, 6, m%10);				
			}
			// ... Stunden
			if (old_h != h) {
				old_h = h;
				// Stunden 10er
				set_area(1, 1, 3, 3, h/10);
				// Stunden 1er
				set_area(4, 1, 6, 3, h%10);				
			}
		}
	}
}
